//
//   omi.c
//
//   Oliver Fromme  <olli@fromme.com>
//   @(#)$Id: omi.c,v 1.43 1998/12/19 03:27:59 olli Exp $
//

static const char cvsid[]
    = "@(#)$Id: omi.c,v 1.43 1998/12/19 03:27:59 olli Exp $";

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <pwd.h>
#include <grp.h>
#include <fcntl.h>

#include "getlopt.h"
#include "omi.h"
#include "cfg.h"
#include "ftp.h"
#include "mir.h"
#include "pro.h"
#include "rex.h"

#if defined(__SunOS_5_5_1) || defined(__SunOS_5_5)
int gethostname(char *name, int namelen);
#endif

#define MINUTES *60

omiconf globalconfig = {
	NULL, NULL, -1,	// pidfile{mask, name, FD}
	NULL, NULL,	// FTP login and email (password)
	NULL,		// FTP account information
	NULL,		// local_addr
	DEBUG_REPORT | DEBUG_STATS | DEBUG_PROGRESS,	// debug flags
	MIRROR_NONE,	// mirror flags
	2 MINUTES,	// timeout_dns
	2 MINUTES,	// timeout_connect
	3 MINUTES,	// timeout_command
	10 MINUTES,	// timeout_read
	10, 10,		// max. # of soft and hard retries
	10, 10,		// max. # of remote and local unrecovered failures
	25, 25, 900,	// re-open delay: start value, increment, maximum
	0,		// deviation
	100,		// removelimit
	0,		// newest
	0755,		// perms_dir
	0644,		// perms_file
	0,		// newgroup
	FALSE,		// changegroup
	FALSE,		// havereporthead
	(simple_list *) NULL,	// remote (mirror) exlude list
	(simple_list *) NULL	// local (delete) exlude list
};

omiconf *cc = &globalconfig;
omipair *cp = NULL;
omisite *ct = NULL;

time_t now;
struct tm start;

char *matchdesc[5] = {
	"no match    ",
	"EXCL (match)",
	"EXCL (deflt)",
	"INCL (match)",
	"INCL (deflt)"
};

void
skipwhitespace (char **cptr) {
	while (**cptr == ' ' || **cptr == '\t')
		(*cptr)++;
}

void
addstrlist (char ***strsref, int *count, char *new)
{
	if (!(*strsref = realloc(*strsref, (*count + 1) * sizeof(**strsref))))
		out_of_memory();
	if (!((*strsref)[*count] = strdup(new)))
		out_of_memory();
	(*count)++;
}

const topt myopts[];

void
listoptions (void)
{
	const topt *o, *cmp;

	cmp = NULL;
	for (o = myopts; o->sname || o->lname; o++)
		if (o->help)
			fprintf (stderr, "   -%c   %s\n", o->sname, o->help);
}

int
usage (char *arg)
{
	fprintf (stderr,
	    "Usage:  %s [package ... | option ...]\n"
	    "This is version " OMI_VERSION " of Olli's mirror \"omi\".\n"
	    "Options (please refer to the documentation for details!):\n", me);
	listoptions();
	exit (1);
	return 0;
}

char **pkgdir = NULL;
int numpkgdirs = 0;

typedef struct {
	char *name;
	unsigned long mask;
	char *help;
} flaginfotype;

const flaginfotype debuginfo[] = {
	{"all",          DEBUG_ALL,		"Everything (HUGE output!)"},
	{"none",         DEBUG_NONE,		"Nothing (except for fatal errors)"},
	{"output",       DEBUG_OUTPUT,		"Output to the FTP server"},
	{"input",        DEBUG_INPUT,		"Input from the FTP server (single lines only)"},
	{"create-dirs",  DEBUG_CREATE_DIRS,	"Creation of directories"},
	{"create-files", DEBUG_CREATE_FILES,	"Creation of plain files"},
	{"create-links", DEBUG_CREATE_LINKS,	"Creation of links"},
	{"remove-dirs",  DEBUG_REMOVE_DIRS,	"Removal of directories"},
	{"remove-files", DEBUG_REMOVE_FILES,	"Removal of plain files"},
	{"remove-links", DEBUG_REMOVE_LINKS,	"Removal of links"},
	{"reasoning",    DEBUG_REASONING,	"Reasons for file transfers"},
	{"retry",        DEBUG_RETRY,		"Information about retries"},
	{"report",       DEBUG_REPORT,		"Standard mirror report"},
	{"stats",        DEBUG_STATS,		"Standard mirror statistics"},
	{"statistics",   DEBUG_STATISTICS,	"Verbose statistics for multiple trees"},
	{"dns",          DEBUG_DNS,		"DNS lookups"},
	{"connect",      DEBUG_CONNECT,		"Connection status"},
	{"xferrate",     DEBUG_XFERRATE,	"Report transferrate for every file"},
	{"multi",        DEBUG_MULTI,		"Multi-line results from server input"},
	{"parse",        DEBUG_PARSE,		"Parse result of the remote directory listing"},
	{"dirs",         DEBUG_DIRS,		"create-dirs  + remove-dirs"},
	{"files",        DEBUG_FILES,		"create-files + remove-files"},
	{"links",        DEBUG_LINKS,		"create-links + remove-links"},
	{"create",       DEBUG_CREATE,		"create-dirs + create-files + create-links"},
	{"remove",       DEBUG_REMOVE,		"remove-dirs + remove-files + remove-links"},
	{"server",       DEBUG_SERVER,		"All server activity"},
	{"mirror",       DEBUG_MIRROR,		"All mirror activity"},
	{"verbose",      DEBUG_VERBOSE,		"server + mirror"},
	{"chx",          DEBUG_CHX,		"Report chmod, chgrp, rename"},
	{"remote-regex", DEBUG_REMOTE_REGEX,	"Matching of remote regular expressions"},
	{"local-regex",  DEBUG_LOCAL_REGEX,	"Matching of local regular expressions"},
	{"regex",        DEBUG_REGEX,		"remote-regex + local-regex"},
	{"progress",     DEBUG_PROGRESS,	"Standard progress display"},
	{NULL, 0, NULL}
};

const flaginfotype mirrorinfo[] = {
	{"none",          MIRROR_NONE,		"Nothing"},
	{"getolderfiles", MIRROR_GETOLDERFILES,	"Update files that are older than the local copies"},
	{"preservedirs",  MIRROR_PRESERVEDIRS,	"Do not kill any local directories"},
	{"ignoretime",    MIRROR_IGNORETIME,	"Ignore all time stamps when comparing files"},
	{"donttouch",     MIRROR_DONTTOUCH,	"Do not update time stamps of local files"},
	{"localtime",     MIRROR_LOCALTIME,	"Assume that the FTP server reports local time (not GMT)"},
	{"excludeall",    MIRROR_EXCLUDEALL,	"Exclude everything by default (use include patters)"},
	{"test",          MIRROR_TEST,		"Do not mirror anything, just \"simulate\" it"},
	{"active",        MIRROR_ACTIVE,	"Use active mode (not passive mode) for file transfers"},
	{"nostat",        MIRROR_NOSTAT,	"Avoid \"STAT\" command (some servers can't handle it)"},
	{"permissions",   MIRROR_PERMISSIONS,	"Copy permissions (r, w, x) from server"},
	{NULL, 0}
};

int
setflags (const flaginfotype info[], unsigned long *value,
    char *flags, const char *name)
{
	char *cptr;
	unsigned long mask;
	int i, num, mlen;
	char prefix;

	mask = *value;
	if (!flags) {
		return 0;
	}
	num = 0;
	skipwhitespace (&flags);
	while (*flags) {
		prefix = *flags;
		if (prefix == '-' || prefix == '+')
			flags++;
		else if (num)
			prefix = '+';
		num++;
		cptr = flags;
		while (*cptr) {
			if (*cptr >= 'A' && *cptr <= 'Z')
				*cptr++ += 'a' - 'A';
			else if (*cptr == '_')
				*cptr++ = '-';
			else if ((*cptr >= 'a' && *cptr <= 'z')
			    || (*cptr >= '0' && *cptr <= '9')
			    || (*cptr == '-'))
				cptr++;
			else
				break;
		}
		if ((mlen = cptr - flags) > 0) {
			if (!strncmp(flags, "list", mlen)
			    || !strncmp(flags, "help", mlen)) {
				fprintf (stderr,
				    "%s: Known %s flags (please refer to "
				    "the documentation for details!):\n",
				    me, name);
				for (i = 0; info[i].name; i++)
					if (info[i].help)
						fprintf (stderr, "   %-15s"
						    " %s\n", info[i].name,
						    info[i].help);
				exit (1);
			}
			for (i = 0; info[i].name; i++) {
				if (!strncmp(flags, info[i].name, mlen)) {
					switch (prefix) {
					case '+':
						mask |= info[i].mask;
						break;
					case '-':
						mask &= ~info[i].mask;
						break;
					default:
						mask = info[i].mask;
					}
					break;
				}
			}
			if (!info[i].name) {
				fprintf (stderr, "%s: unknown %s flag "
				    "\"%s\" (try flag \"help\")\n",
				    me, name, strndup(flags, mlen));
				return -1;
			}
		}
		if (*cptr) {
			flags = cptr + 1;
			skipwhitespace (&flags);
		}
		else
			break;
	}
	*value = mask;
	return 0;
}

int
setdebuglevel (char *lvl)
{
	return setflags(debuginfo, &cc->debug, lvl, "debug");
}

int
setmirrorflags (char *mflags)
{
	return setflags(mirrorinfo, &cc->flags, mflags, "mirror");
}

long getport (char *s)
{
	long port;
	char *eptr;

	if (!s || !*s || *s == '-')
		return DEFAULT_FTP_PORT;
	port = strtoul(s, &eptr, 10);
	if (*eptr) {
		fprintf (stderr, "%s: syntax error in port number \"%s\"\n",
		    me, s);
		return -1;
	}
	return port;
}

int
addpkgdir (char *dir)
{
	if (!dir || !*dir) {
		fprintf (stderr,
		    "%s: missing argument (directory name)\n", me);
		return -1;
	}
	addstrlist (&pkgdir, &numpkgdirs, dir);
	return 0;
}

int
loadpackage (char *name)
{
	static char **pkghistory = NULL;
	static int pkghistsize = 0;
	char *path;
	int result, i;

	if (!name || !*name) {
		fprintf (stderr, "%s: missing argument (file name)\n", me);
		return -1;
	}
	path = NULL;
	if (!strchr(name, '/') && numpkgdirs) {
		struct stat st;
		int count, nlen;

		nlen = strlen(name);
		for (count = numpkgdirs - 1; count >= 0; count--) {
			if (path)
				free (path);
			path = tmalloc(strlen(pkgdir[count]) + nlen + 2);
			strcpy (path, pkgdir[count]);
			strcat (path, "/");
			strcat (path, name);
			if (stat(path, &st) >= 0) {
				name = path;
				break;
			}
		}
	}
	for (i = 0; i < pkghistsize; i++)
		if (!strcmp(name, pkghistory[i])) {
			if (path)
				free (path);
			return 0;
		}
	result = cfg_parsefile(myopts, name, 0);
	addstrlist (&pkghistory, &pkghistsize, name);
	if (path)
		free (path);
	return result;
}

omisite *site = NULL;
int numsites = 0;
int lastsite = 0;

int
addserver (char *srv)
{
	omisite *os;
	char *sep;
	unsigned int p;
	int i;

	p = DEFAULT_FTP_PORT;
	if (srv && ((sep = strrchr(srv, ':')) || (sep = strrchr(srv, ' ')))) {
		long tp;

		*sep++ = '\0';
		if ((tp = getport(sep)) <= -1)
			return tp;
		p = tp;
	}
	if (!srv || !*srv) {
		fprintf (stderr, "%s: missing argument (server name)\n", me);
		return -1;
	}
	for (i = 0; i < numsites; i++)
		if (!strcmp(site[i].name, srv) && site[i].port == p) {
			lastsite = i;
			cc = site[i].conf;
			return 0;
		}
	if (!(site = realloc(site, (numsites + 1) * sizeof(*site))))
		out_of_memory();
	os = site + (lastsite = numsites++);
	if (!(os->name = strdup(srv)))
		out_of_memory();

	os->port = p;
	os->conf = cc = tmalloc(sizeof(omiconf));
	memcpy (cc, &globalconfig, sizeof(omiconf));
	if (globalconfig.pidfilemask) {
		if (!(cc->pidfilemask = strdup(globalconfig.pidfilemask)))
			out_of_memory();
	}
	else
		cc->pidfilemask = NULL;
	if (globalconfig.local_addr) {
		if (!(cc->local_addr = strdup(globalconfig.local_addr)))
			out_of_memory();
	}
	else
		cc->local_addr = NULL;
	if (!(cc->login = strdup(globalconfig.login)) ||
	    !(cc->email = strdup(globalconfig.email)))
		out_of_memory();
	if (globalconfig.account &&
	    !(cc->account = strdup(globalconfig.account)))
		out_of_memory();
	cc->exclude = NULL;
	cc->locexclude = NULL;

	os->stats = tmalloc(sizeof(omistats));
	//   Don't initialize it here, but rather when it's
	//   used for the 1st time, so the start time is set
	//   correctly!

	os->num = 0;
	os->pair = NULL;
	return 0;
}

int
addremotedir (char *dir)
{
	omisite *os;
	omipair *op;
	omifile *od;

	if (!dir || !*dir) {
		fprintf (stderr,
		    "%s: missing argument (remote directory name)\n", me);
		return -1;
	}
	if (lastsite >= numsites) {
		fprintf (stderr,
		    "%s: remote server must be specified first\n", me);
		return -1;
	}
	os = site + lastsite;
	if (!(os->pair = realloc(os->pair, (os->num + 1) * sizeof(omipair))))
		out_of_memory();
	op = os->pair + os->num++;
	if (!(op->localdir = strdup(dir)))
		out_of_memory();
	if (!(op->remotedir = strdup(dir)))
		out_of_memory();
	od = tmalloc(sizeof(*op->omidir));
	op->omidir = od;
	od->mode = S_IFDIR | os->conf->perms_dir;
	od->size = 0;
	od->mtime = 0;	//   don't update mtime of local file
	od->tprec = 0;
	od->name = "";
	od->data = NULL;

	op->conf = cc = tmalloc(sizeof(*op->conf));
	memcpy (op->conf, os->conf, sizeof(omiconf));
	if (os->conf->pidfilemask) {
		if (!(cc->pidfilemask = strdup(os->conf->pidfilemask)))
			out_of_memory();
	}
	else
		cc->pidfilemask = NULL;
	if (os->conf->local_addr) {
		if (!(cc->local_addr = strdup(os->conf->local_addr)))
			out_of_memory();
	}
	else
		cc->local_addr = NULL;
	if (!(cc->login = strdup(os->conf->login)) ||
	    !(cc->email = strdup(os->conf->email)))
		out_of_memory();
	if (os->conf->account &&
	    !(cc->account = strdup(os->conf->account)))
		out_of_memory();
	cc->exclude = NULL;
	cc->locexclude = NULL;

	op->stats = tmalloc(sizeof(omistats));
	//   Don't initialize it here, but rather when it's
	//   used for the 1st time, so the start time is set
	//   correctly!

	return 0;
}

int
setlocaldir (char *dir)
{
	omisite *os;
	omipair *op;

	if (!dir || !*dir) {
		fprintf (stderr,
		    "%s: missing argument (local directory name)\n", me);
		return -1;
	}
	if (lastsite >= numsites) {
		fprintf (stderr,
		    "%s: remote server must be specified first\n", me);
		return -1;
	}
	os = site + lastsite;
	if (!os->num) {
		fprintf (stderr,
		    "%s: remote directory must be specified first\n", me);
		return -1;
	}
	op = os->pair + (os->num - 1);
	if (op->localdir)
		free (op->localdir);
	if (!(op->localdir = strdup(dir)))
		out_of_memory();
	return 0;
}

int
setlocalbind (char *locaddr)
{
	if (cc->local_addr)
		free(cc->local_addr);
	if (!locaddr || !*locaddr || *locaddr == '-') {
		cc->local_addr = NULL;
		return 0;
	}
	if (!(cc->local_addr = strdup(locaddr)))
		out_of_memory();
	return 0;
}

int
setchgroup (char *groupspec)
{
	char *eptr;

	if (!groupspec || !*groupspec || *groupspec == '-') {
		cc->changegroup = FALSE;
		return 0;
	}
	cc->changegroup = TRUE;
	cc->newgroup = strtoul(groupspec, &eptr, 10);
	if (*eptr) {
		struct group *grp;

		if (!(grp = getgrnam(groupspec))) {
			fprintf (stderr, "%s: unknown group name \"%s\"\n",
			    me, groupspec);
			return -1;
		}
		cc->newgroup = grp->gr_gid;
	}
	return 0;
}

int
setpermode (char *modespec)
{
	char *cptr, *eptr;
	int numpar;

	if (!modespec || !*modespec || !strcmp(modespec, "-")) {
		cc->perms_file = 0644;
		cc->perms_dir  = 0755;
		return 0;
	}
	if ((numpar = csv_count(modespec)) > 2) {
		fprintf (stderr, "%s: syntax error in permission modes "
		    "(more than 2 arguments)\n", me);
		return -1;
	}
	cptr = csv_value(modespec, 0);
	if (*cptr) {
		cc->perms_file = strtol(cptr, &eptr, 8);
		if (*eptr) {
			fprintf (stderr, "%s: syntax error in file "
			    "permission mode \"%s\"\n", me, cptr);
			return -1;
		}
	}
	if (numpar > 1) {
		cptr = csv_value(modespec, 1);
		if (*cptr) {
			cc->perms_dir = strtol(cptr, &eptr, 8);
			if (*eptr) {
				fprintf (stderr, "%s: syntax error in "
				    "directory permission mode \"%s\"\n",
				    me, cptr);
				return -1;
			}
		}
	}
	else {
		cc->perms_dir = cc->perms_file | (cc->perms_file & 0444) >> 2;
		cc->perms_file &= 0666;
	}
	return 0;
}

long
gettimespec (char *ts)
{
	long t, tt;
	char *eptr, c;
	int sub;

	skipwhitespace (&ts);
	if (!*ts) {
		fprintf (stderr, "%s: missing time specification\n", me);
		return -1;
	}
	t = 0;
	sub = FALSE;
	do {
		tt = strtoul(ts, &eptr, 10);
		if (eptr == ts) {
			fprintf (stderr,
			    "%s: syntax error in time specification "
			    "starting with \"%s\"\n", me, ts);
			return -1;
		}
		ts = eptr;
		skipwhitespace (&ts);
		c = *ts;
		if (c == 's' || c == 'S')
			ts++;
		else if (c == 'm' || c == 'M') {
			tt *= 60;	// minutes
			ts++;
		}
		else if (c == 'h' || c == 'H') {
			tt *= 3600;	// hours
			ts++;
		}
		else if (c == 'd' || c == 'D') {
			tt *= 86400;	// days
			ts++;
		}
		else if (c == 'w' || c == 'W') {
			tt *= 604800;	// weeks
			ts++;
		}
		else if (c == 'y' || c == 'Y') {
			tt *= 31556952;	// years
			ts++;
		}
		skipwhitespace (&ts);
		if (sub) {
			t -= tt;
			sub = FALSE;
		}
		else
			t += tt;
		if (*ts == '+')
			ts++;
		else if (*ts == '-') {
			sub = TRUE;
			ts++;
		}
		skipwhitespace (&ts);
	} while (*ts);
	return t;
}

int
setdeviation (char *dev)
{
	long d;

	if (!dev || !*dev) {
		fprintf (stderr,
		    "%s: missing argument (maximum deviation in seconds)\n",
		    me);
		return -1;
	}
	d = gettimespec(dev);
	if (d < 0)
		return -1;
	cc->deviation = d;
	return 0;
}

int
setremovelimit (char *lim)
{
	char *eptr;

	if (!lim || !*lim) {
		fprintf (stderr,
		    "%s: missing argument (remove limit in percent)\n", me);
		return -1;
	}
	cc->removelimit = strtol(lim, &eptr, 10);
	if (*eptr) {
		fprintf (stderr, "%s: illegal integer value\n", me);
		return -1;
	}
	if (cc->removelimit < 0 || cc->removelimit > 100) {
		fprintf (stderr, "%s: illegal remove limit "
		    "(must be a percent value in the range 0...100)\n", me);
		return -1;
	}
	return 0;
}

int
setnewest (char *n)
{
	char *eptr;

	if (!n || !*n) {
		fprintf (stderr, "%s: missing argument "
		    "(number of newest files to mirror)\n", me);
		return -1;
	}
	cc->newest = strtol(n, &eptr, 10);
	if (*eptr) {
		fprintf (stderr, "%s: illegal integer value\n", me);
		return -1;
	}
	if (cc->newest < 0) {
		fprintf (stderr, "%s: illegal number (must be >= 0)\n", me);
		return -1;
	}
	return 0;
}

int
setclude (char *clude, simple_list **l, unsigned int flags, char *what)
{
	if (!clude || !*clude) {
		fprintf (stderr, "%s: missing argument "
		    "(regular expression of %sclude list)\n", me, what);
		return -1;
	}
	if (!*l) {
		*l = tmalloc(sizeof(simple_list));
		simple_list_init (*l);
	}
	if (!add_regex(*l, clude, flags)) {
		fprintf (stderr,
		    "%s: error in regular expression of %sclude list\n",
		    me, what);
		return -1;
	}
	return 0;
}

int
setexclude (char *excl)
{
	return setclude(excl, &cc->exclude, 0, "remote ex");
}

int
setlocexclude (char *excl)
{
	return setclude(excl, &cc->locexclude, 0, "local ex");
}

int
setinclude (char *incl)
{
	return setclude(incl, &cc->exclude, RE_INCLUDE, "remote in");
}

int
setlocinclude (char *incl)
{
	return setclude(incl, &cc->locexclude, RE_INCLUDE, "local in");
}

int
setpidfile (char *name)
{
	if (cc->pidfilemask)
		free (cc->pidfilemask);
	if (!name || !*name)
		cc->pidfilemask = NULL;
	else
		if (!(cc->pidfilemask = strdup(name)))
			out_of_memory();
	return 0;
}

int
setloginname (char *name)
{
	if (!name || !*name) {
		fprintf (stderr,
		    "%s: missing argument (login / user name)\n", me);
		return -1;
	}
	if (cc->login)
		free (cc->login);
	if (!(cc->login = strdup(name)))
		out_of_memory();
	return 0;
}

int
setemail (char *name)
{
	if (!name || !*name) {
		fprintf (stderr,
		    "%s: missing argument (email address / password)\n", me);
		return -1;
	}
	if (cc->email)
		free (cc->email);
	if (!(cc->email = strdup(name)))
		out_of_memory();
	return 0;
}

int
setaccount (char *name)
{
	if (!name || !*name) {
		fprintf (stderr,
		    "%s: missing argument (account information)\n", me);
		return -1;
	}
	if (cc->account)
		free (cc->account);
	if (!(cc->account = strdup(name)))
		out_of_memory();
	return 0;
}

typedef struct {
	char tag;
	unsigned int *var;
} taglist;

int
gettaglist (char *list, taglist *tagl, char *description, bool istime)
{
	int i, j;
	long value;
	char *cptr, *eptr;
	char type;

	if (!list || !*list) {
		fprintf (stderr,
		    "%s: missing argument (%s specification(s))\n",
		        me, description);
		return -1;
	}
	i = 0;
	while ((cptr = csv_value(list, i++))) {
		if (!(type = *cptr)) {
			free (cptr);
			continue;
		}
		if (type >= 'A' && type <= 'Z')
			type += 'a' - 'A';
		if (istime)
			value = gettimespec(cptr + 1);
		else {
			value = strtoul(cptr + 1, &eptr, 10);
			if (*eptr) {
				fprintf (stderr,
				    "%s: syntax error in %s number (\"%s\")\n",
				    me, description, cptr + 1);
				return -1;
			}
		}
		free (cptr);
		if (value < 0)
			return -1;
		for (j = 0; tagl[j].var; j++)
			if (type == tagl[j].tag) {
				*tagl[j].var = value;
				break;
			}
		if (!tagl[j].var) {
			fprintf (stderr,
			    "%s: unknown %s type letter \"%c\"\n",
			    me, description, type);
			return -1;
		}
	}
	return 0;
}

taglist timeout_tags[] = {
	{'d', NULL},
	{'n', NULL},
	{'m', NULL},
	{'r', NULL},
	{0, NULL}
};

int
settimeout (char *list)
{
	timeout_tags[0].var = &cc->timeout_dns;
	timeout_tags[1].var = &cc->timeout_connect;
	timeout_tags[2].var = &cc->timeout_command;
	timeout_tags[3].var = &cc->timeout_read;
	return gettaglist(list, timeout_tags, "timeout", TRUE);
}

taglist retry_tags[] = {
	{'s', NULL},
	{'h', NULL},
	{'f', NULL},
	{'l', NULL},
	{0, NULL}
};

int
setretry (char *list)
{
	retry_tags[0].var = &cc->maxsretry;
	retry_tags[1].var = &cc->maxhretry;
	retry_tags[2].var = &cc->maxremfail;
	retry_tags[3].var = &cc->maxlocfail;
	return gettaglist(list, retry_tags, "retry", FALSE);
}

taglist delay_tags[] = {
	{'d', NULL},
	{'i', NULL},
	{'m', NULL},
	{0, NULL}
};

int
setdelay (char *list)
{
	delay_tags[0].var = &cc->reopendelay;
	delay_tags[1].var = &cc->reopeninc;
	delay_tags[2].var = &cc->reopenmax;
	return gettaglist(list, delay_tags, "reopen delay", TRUE);
}

const topt myopts[] = {
	{'s', "server",        GLO_CHAR, addserver,     0, 0, "Remote server name or IP (can be followed by \":<portnumber>\")"},
	{'r', "remotedir",     GLO_CHAR, addremotedir,  0, 0, "Remote directory"},
	{'l', "localdir",      GLO_CHAR, setlocaldir,   0, 0, "Local directory"},
	{'b', "bind",          GLO_CHAR, setlocalbind,  0, 0, "Bind local address"},
	{'d', "debug",         GLO_CHAR, setdebuglevel, 0, 0, "Debug & verbosity options (\"-d help\" for a list)"},
	{'v', "verbosity",     GLO_CHAR, setdebuglevel, 0, 0, NULL},
	{'o', "options",       GLO_CHAR, setmirrorflags,0, 0, "Mirror options (\"-o help\" for a list)"},
	{'p', "pidfile",       GLO_CHAR, setpidfile,    0, 0, "PID file / lock file"},
	{'p', "lockfile",      GLO_CHAR, setpidfile,    0, 0, NULL},
	{'g', "group",         GLO_CHAR, setchgroup,    0, 0, "Group to chgrp files (name or numeric GID)"},
	{'m', "mode",          GLO_CHAR, setpermode,    0, 0, "Permission modes to chmod files (octal)"},
	{'t', "deviation",     GLO_CHAR, setdeviation,  0, 0, "Maximum acceptable time deviation"},
	{'P', "removelimit",   GLO_CHAR, setremovelimit,0, 0, "Maximum remove limit (percent), default: 100"},
	{'n', "newest",        GLO_CHAR, setnewest,     0, 0, "Only mirror the <n> newest files in each dir"},
	{'x', "exclude",       GLO_CHAR, setexclude,    0, 0, "Exclude pattern for remote files (reg.exp.)"},
	{'x', "remoteexclude", GLO_CHAR, setexclude,    0, 0, NULL},
	{'X', "localexclude",  GLO_CHAR, setlocexclude, 0, 0, "Exclude pattern for local files (reg.exp.)"},
	{'y', "include",       GLO_CHAR, setinclude,    0, 0, "Include pattern for remote files (reg.exp.)"},
	{'y', "remoteinclude", GLO_CHAR, setinclude,    0, 0, NULL},
	{'Y', "localinclude",  GLO_CHAR, setlocinclude, 0, 0, "Include pattern for local files (reg.exp.)"},
	{'u', "login",         GLO_CHAR, setloginname,  0, 0, "User name for FTP login (default: \"ftp\")"},
	{'u', "user",          GLO_CHAR, setloginname,  0, 0, NULL},
	{'e', "email",         GLO_CHAR, setemail,      0, 0, "Email address for FTP login (default: \"<your_login>@\")"},
	{'a', "account",       GLO_CHAR, setaccount,    0, 0, "Account information for FTP login (default: none)"},
	{'T', "timeout",       GLO_CHAR, settimeout,    0, 0, "Timeout specifications"},
	{'R', "retry",         GLO_CHAR, setretry,      0, 0, "Retry count specifications"},
	{'D', "delay",         GLO_CHAR, setdelay,      0, 0, "Retry delay specifications"},
	{'P', "packagedir",    GLO_CHAR, addpkgdir,     0, 0, NULL},
	{'I', "includedir",    GLO_CHAR, addpkgdir,     0, 0, "Directory for include / package files"},
	{'i', "package",       GLO_CHAR, loadpackage,   0, 0, "Load include / package file"},
	{'i', "include",       GLO_CHAR, loadpackage,   0, 0, NULL},
	{'h', "help", GLO_NONE, usage, 0, 0, "Print this help"},
	{'?', 0,      GLO_NONE, usage, 0, 0, NULL},
	{0, 0, 0, 0, 0, 0, 0}
};

static int
lock_file (int fd)
{
	struct flock fl;

	fl.l_type = F_WRLCK;
	fl.l_whence = SEEK_SET;
	fl.l_start = fl.l_len = 0;
	return fcntl(fd, F_SETLK, &fl);
}

static char *
pidexpand (char *mask)
{
	char *pn;

	/* return string_expand(mask); */
	if (!(pn = strdup(mask)))
		out_of_memory();
	return pn;
}

static bool
create_lockfile (char *globalfile, char *serverfile)
{
	char buf[32];

	if (!cc->pidfilemask) {
		cc->pidfilename = NULL;
		return TRUE;
	}
	cc->pidfilename = pidexpand(cc->pidfilemask);
	if (globalfile) {
		if (!strcmp(cc->pidfilename, globalfile) ||
		    (serverfile && !strcmp(cc->pidfilename, serverfile))) {
			free (cc->pidfilename);
			cc->pidfilename = NULL;
			return TRUE;
		}
	}
	cc->pidfiledesc = open(cc->pidfilename, O_RDWR | O_CREAT, 0664);
	if (cc->pidfiledesc < 0) {
		fprintf (stderr, "%s: can't create lockfile:\n", me);
		fprintf (stderr, "%s: %s\n", cc->pidfilename, serror());
		free (cc->pidfilename);
		cc->pidfilename = NULL;
		return FALSE;
	}
	if (lock_file(cc->pidfiledesc) < 0) {
		close (cc->pidfiledesc);
		fprintf (stderr, "%s: can't lock lockfile:\n", me);
		fprintf (stderr, "%s: %s\n", cc->pidfilename, serror());
		free (cc->pidfilename);
		cc->pidfilename = NULL;
		return FALSE;
	}
	sprintf (buf, "%d\n", getpid());
	write (cc->pidfiledesc, buf, strlen(buf) + 1);
	return TRUE;
}

static void
remove_lockfile (void)
{
	if (!cc->pidfilename)
		return;
	close (cc->pidfiledesc);
	unlink (cc->pidfilename);
	free (cc->pidfilename);
	cc->pidfilename = NULL;
}

static void
init (void)
{
	struct passwd *pw;

	stats_init (&globalstats);
	now = globalstats.start;
	memcpy (&start, localtime(&now), sizeof(struct tm));
	globalconfig.pidfilemask = NULL;
	if (!(globalconfig.login = strdup("ftp")))
		out_of_memory();
	if ((pw = getpwuid(getuid())) && pw->pw_name) {
		char buff[1024];

		memset (buff, 0, 1024);
		strncpy (buff, pw->pw_name, 1020);
		strcat (buff, "@");
		if (!(globalconfig.email = strdup(buff)))
			out_of_memory();
	}
	else if (!(globalconfig.email = strdup("omi@")))
		out_of_memory();
	init_progress();
}

static void
done (void)
{
	done_progress();
}

static ftpreply *
getstat (ftpstat *fs, char *dir)
{
	ftpreply *fr;

	fr = ftp_needtype(fs, TYPE_ASCII);
	if (fr->code >= 400)
		return fr;
	ftp_freereply (fr);
	fr = ftp_needcwd(fs, dir);
	if (fr->code >= 400)
		return fr;
	ftp_freereply (fr);
	return ftp_cmd(fs, "STAT -lRa");
}

static int
actual_mirror_tree (omipair *op, ftpstat *fs)
{
	ftpreply *myreply;
	char *cptr;
	int result;

	result = 0;
	if (!(cc->flags & MIRROR_NOSTAT)) {
		if (fs->support & FTPD_STAT)
			myreply = ftp_sequence(fs, (sequencejob *) getstat,
			    op->remotedir);
		else
			myreply = ftp_fakereply(499, RETRY_NO,
			    "STAT command not supported");
		if (myreply->code < 400) {
			//   Skip the "213-status of ..." line:
			cptr = myreply->text + strcspn(myreply->text, "\r\n");
			cptr += strspn(cptr, "\r\n");
			result = parse_lslr(cptr, op->omidir, op->remotedir);
		}
		ftp_freereply (myreply);
	}
	if (!result) {
		//
		//   The STAT command did not work for some
		//   reason (not supported, or RFC-ignorant),
		//   so we try LIST instead.
		//
		fs->support &= ~(unsigned) FTPD_STAT;
		myreply = ftp_lslra(fs, op->remotedir, &cptr);
		if (myreply->code >= 400) {
			fprintf (stderr,
			    "%s: can't read remote directory\n", me);
			result = myreply->code;
			ftp_freereply (myreply);
			return -1;
		}
		if (cc->debug & DEBUG_MULTI)
			fputs (cptr, stderr);
		result = parse_lslr(cptr, op->omidir, op->remotedir);
		ftp_freereply (myreply);
		free (cptr);
	}
	if (cc->debug & DEBUG_PARSE)
		debug_tree (op->remotedir, op->omidir);
	if (mkdir_parents(op->localdir) >= 0) {
		op->omidir->mode = S_IFDIR | cc->perms_dir;
		result = mirror_dir(fs, op->remotedir, op->localdir,
		    op->omidir);
	}
	else {
		fprintf (stderr, "%s: Can't create local directory, "
		    "mirror aborted.\n", me);
		result = -1;
	}
	free_tree (op->omidir);
	return result;
}

static int
mirror_tree (int j, omisite *os, ftpstat *fs)
{
	omipair *op;
	int result;

	op = cp = os->pair + j;
	cc = op->conf;
	cs = op->stats;
	stats_init (cs);
	result = 0;
	if (create_lockfile(globalconfig.pidfilename, os->conf->pidfilename))
		result = actual_mirror_tree (op, fs);
	remove_lockfile();
	stats_done (cs);
	if (cc->debug & DEBUG_STATISTICS)
		stats_report (cs, os->name,
		    op->remotedir, op->localdir);
	stats_summary (os->stats, cs);
	os->stats->totalpairs++;
	return result;
}

static void
mirror_site (int i)
{
	ftpstat *myftp;
	omisite *os;
	int j, result;

	os = ct = site + i;
	cc = os->conf;
	cs = os->stats;
	stats_init (cs);
	if (create_lockfile(globalconfig.pidfilename, NULL)) {
		myftp = ftp_initserver(os->name, os->port);
		for (j = 0; j < os->num; j++) {
			result = mirror_tree(j, os, myftp);
			if (result <= -3)
				break;
		}
		ftp_freereply (ftp_doneserver(myftp));
	}
	cc = os->conf;
	remove_lockfile();
	cs = os->stats;
	stats_done (cs);
	if (cc->debug & DEBUG_STATISTICS && cs->totalpairs > 1)
		stats_report (cs, os->name, NULL, NULL);
	stats_summary (&globalstats, cs);
	globalstats.totalsites++;
}

int
main (int argc, char *argv[])
{
	int i;

	utils_init (argv[0]);
	if (!geteuid() || !getuid())
		fprintf (stderr, "%s: WARNING: running as root!\n", me);
	init();
	if (cfg_globalrcfile(myopts, 1) < 0)
		return 1;
	if (cfg_localrcfile(myopts, 1) < 0)
		return 1;
	do {
		if (parselopts(argc, argv, myopts, me) < 0)
			return 1;
		if (loptind < argc)
			if (loadpackage(argv[loptind++]) < 0)
				return 1;
	} while (loptind < argc);

	cc = &globalconfig;
	if (create_lockfile(NULL, NULL))
		for (i = 0; i < numsites; i++)
			mirror_site (i);
	cc = &globalconfig;
	remove_lockfile();
	cs = &globalstats;
	stats_done (cs);
	if ((cc->debug & DEBUG_STATISTICS && cs->totalsites > 1) ||
	    (cc->debug & DEBUG_STATS && cs->totalchanges > 0))
		stats_report (cs, NULL, NULL, NULL);
	done();

	return 0;
}

//-- 
