//
//   stt.c
//
//   Oliver Fromme  <olli@fromme.com>
//   @(#)$Id: stt.c,v 1.15 1998/12/11 02:11:02 olli Exp $
//

static const char cvsid[]
    = "@(#)$Id: stt.c,v 1.15 1998/12/11 02:11:02 olli Exp $";

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <time.h>
#include <fcntl.h>

#include "stt.h"

omistats globalstats;
omistats *cs = &globalstats;

void
stats_init (omistats *os)
{
	memset (os, 0, sizeof(omistats));
	os->start = os->end = time(NULL);
}

void
stats_done (omistats *os)
{
	os->totalchanges =
	    os->gonefiles + os->gonedirs + os->gonelinks +
	    os->newfiles  + os->newdirs  + os->newlinks +
	    os->updfiles  + os->upddirs  + os->updlinks +
	    os->permfiles + os->permdirs;
	os->end = time(NULL);
}

void
stats_summary (omistats *sum, omistats *x)
{
	sum->errors	+= x->errors;
	sum->totalpairs	+= x->totalpairs;
	sum->totalbytes	+= x->totalbytes;
	sum->totalfiles	+= x->totalfiles;
	sum->totaldirs	+= x->totaldirs;
	sum->totallinks	+= x->totallinks;
	sum->gonebytes	+= x->gonebytes;
	sum->gonefiles	+= x->gonefiles;
	sum->gonedirs	+= x->gonedirs;
	sum->gonelinks	+= x->gonelinks;
	sum->newbytes	+= x->newbytes;
	sum->newfiles	+= x->newfiles;
	sum->newdirs	+= x->newdirs;
	sum->newlinks	+= x->newlinks;
	sum->updbytes	+= x->updbytes;
	sum->updfiles	+= x->updfiles;
	sum->upddirs	+= x->upddirs;
	sum->updlinks	+= x->updlinks;
	sum->permfiles	+= x->permfiles;
	sum->permdirs	+= x->permdirs;
	sum->growbytes	+= x->growbytes;
	sum->xferbytes	+= x->xferbytes;
	sum->xfertime	+= x->xfertime;
}

static char *
maketimespec (time_t t)
{
	static char buffer[40];
	struct tm *timeptr;

	timeptr = localtime(&t);
	strftime (buffer, 40, "%e-%b-%Y %H:%M:%S", timeptr);
	return buffer;
}

char *
ulkformat (unsigned long ulk)
{
	static char buf[32];
	int commas, digits;

	sprintf (buf, "%lu", ulk);
	digits = strlen(buf);
	if ((commas = (digits - 1) / 3)) {
		char *src, *dst;
		int i;

		src = buf + (digits - 1);
		dst = src + commas;
		dst[1] = '\0';
		i = 0;
		while (++i <= digits) {
			*dst-- = *src--;
			if (!(i % 3))
				*dst-- = ',';
		}
	}
	return buf;
}

char *
lkformat (long lk)
{
	static char buf[32];
	char *vbuf;
	int commas, digits;

	sprintf (buf, "%ld", lk);
	if (*buf == '-')
		vbuf = buf + 1;
	else
		vbuf = buf;
	digits = strlen(vbuf);
	if ((commas = (digits - 1) / 3)) {
		char *src, *dst;
		int i;

		src = vbuf + (digits - 1);
		dst = src + commas;
		dst[1] = '\0';
		for (i = 0; i < digits; i++) {
			if (!(i % 3) && i)
				*dst-- = ',';
			*dst-- = *src--;
		}
	}
	return buf;
}

char *
xferrate_report (double duration, ulonglong bytes)
{
	static char buf[32];
	double rate;

	if (duration >= 0.0001 && bytes) {
		rate = bytes / duration;
		if (rate < 1000)
			sprintf (buf, "%6.2f bytes/sec", rate);
		else if (rate < 1024000)
			sprintf (buf, "%6.2f Kbytes/sec", rate / 1024);
		else
			sprintf (buf, "%6.2f Mbytes/sec", rate / 1048576);
	}
	else
		strcpy (buf, "n/a");
	return buf;
}

void
stats_report (omistats *os, char *site, char *remote, char *local)
{
	putc ('\n', stderr);
	if (remote && local)
		fprintf (stderr, "Statistics for %s:%s -> %s\n",
		    site, remote, local);
	else if (site)
		if (os->totalpairs != 1)
			fprintf (stderr,
			    "Statistics for %s (%u directory trees)\n",
			    site, os->totalpairs);
		else
			fprintf (stderr, "Statistics for %s\n", site);
	else {
		fprintf (stderr, "Statistics summary");
		if (os->totalsites != 1 || os->totalpairs != 1)
			fprintf (stderr,
			    " (%u servers, %u directory trees)\n",
			    os->totalsites, os->totalpairs);
		else
			putc ('\n', stderr);
	}
	fprintf (stderr, "start: %s", maketimespec(os->start));
	fprintf (stderr, ", end: %s", maketimespec(os->end));
	fprintf (stderr, ", duration: %s\n", time_format(difftime(os->end, os->start)));
	fprintf (stderr, "Average transfer rate:  %s\n",
	    xferrate_report(os->xfertime, os->xferbytes));
	fprintf (stderr, "Number of unrecovered errors occured:  %u\n\n",
	    os->errors);
	fprintf (stderr, "                     |     Kbytes  |  files  |   dirs  |  links\n");
	fprintf (stderr, "  -------------------+-------------+---------+---------+---------\n");
	fprintf (stderr, "    total in mirror  |%11s  |%7u  |%7u  |%7u\n", ulkformat((os->totalbytes + 512) / 1024), os->totalfiles, os->totaldirs, os->totallinks);
	fprintf (stderr, "    removed          |%11s  |%7u  |%7u  |%7u\n", ulkformat((os->gonebytes + 512) / 1024), os->gonefiles, os->gonedirs, os->gonelinks);
	fprintf (stderr, "    updated          |%11s  |%7u  |%7u  |%7u\n", ulkformat((os->updbytes + 512) / 1024), os->updfiles, os->upddirs, os->updlinks);
	fprintf (stderr, "    created          |%11s  |%7u  |%7u  |%7u\n", ulkformat((os->newbytes + 512) / 1024), os->newfiles, os->newdirs, os->newlinks);
	fprintf (stderr, "    permissions      |%11s  |%7u  |%7u  |%7u\n", "n/a", os->permfiles, os->permdirs, 0);
	fprintf (stderr, "    growth           |%11s  |%7d  |%7d  |%7d\n", lkformat((os->growbytes + 512) / 1024), (int) os->newfiles - (int) os->gonefiles, (int) os->newdirs - (int) os->gonedirs, (int) os->newlinks - (int) os->gonelinks);
	putc ('\n', stderr);
}

//-- 
