/*
 *   utils.c
 *
 *   Oliver Fromme  <olli@fromme.com>
 *   @(#)$Id: utils.c,v 1.3 1998/02/09 17:15:48 inof Exp $
 */

static const char cvsid[]
    = "@(#)$Id: utils.c,v 1.3 1998/02/09 17:15:48 inof Exp $";

#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "utils.h"

char *me;

void out_of_memory (void)
{
	fprintf (stderr, "%s: Out of memory.\n", me);
	exit (1);
}

void *tmalloc (size_t size)
{
	void *mem;

	if (!(mem = malloc(size)))
		out_of_memory();
	return (mem);
}

char *strndup (const char *src, int num)
{
	char *dst;

	if (!(dst = (char *) malloc(num+1)))
		return (NULL);
	dst[num] = '\0';
	return (strncpy(dst, src, num));
}

char *justify (char *str)
{
	int si = 0, di = 0, havespc = TRUE;
	char c;

	if (!str)
		return (str);
	/*LINTED*/
	while ((c = str[si])) {
		if (c == ' ' || c == '\t' || c == '\r' || c == '\n')
			if (havespc) {
				si++;
				continue;
			}
			else
				havespc = TRUE;
		else
			havespc = FALSE;
		str[di++] = str[si++];
	}
	if (di && havespc)
		di --;
	str[di] = '\0';
	return (str);
}

char *serror (void)
{
	static char buf[32];
	char *errstr;

	if (!(errstr = strerror(errno))) {
		sprintf (buf, "Unknown error #%d.", errno);
		return buf;
	}
	return errstr;
}

char *me;

void die (const char *func, const char *file)
{
	fprintf (stderr, "%s: ", me);
	if (func)
		fprintf (stderr, "%s(%s): ", func, file?file:"");
	else if (file)
		fprintf (stderr, "%s: ", file);
	fprintf (stderr, "%s\n", serror());
	exit (1);
}

void utils_init (const char *argv0)
{
	if (argv0)
		(me = strrchr(argv0, '/')) ? me++ : (me = (char *) argv0);
	else
		me = "";
}

/* EOF */
